#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
simple_scalar_sin_without_mape_r2.py

This script compares two deterministic I–V models for a tunnel diode:

1) A Quantum Baseline (Shockley-like),
2) A Scalar Field Theory Model, where key parameters
   (Planck length, mass parameter m, and amplitude phi_0)
   are derived deterministically from fundamental constants
   (Planck constant ħ, G, c). No parameter fitting is performed.

We only compute RMS, MAE, AIC, and BIC, omitting MAPE and R²
(because certain data regimes caused large percentage errors
or negative R²).

Data Source:
------------
"TunneldiodeTotal.csv" from Kaggle:
https://www.kaggle.com/datasets/raktim711/experimental-data-on-tunnel-diode

Author: Your Name
(c) 2025

--------------------------------------------------------------------------------
THEORY OF THE DETERMINISTIC SCALAR FIELD (Commentary)
--------------------------------------------------------------------------------
1) Deterministic Scalar Field Parameters:
   - According to your theory, the scalar field is characterized by fundamental
     constants: Planck length ℓₚ = sqrt( (ħ·G)/c³ ), a dimensionless factor λ,
     a "mass" parameter m = 1/(λ·ℓₚ), and an amplitude φ₀ = ħ/(λ·ℓₚ·c).
   - In this script:
       lambda_factor = 1e10  (set as a fixed number, not fitted),
       m_val = 1/(lambda_factor * ℓₚ),
       phi_0 = ħ/(lambda_factor * ℓₚ·c).
   - These yield extremely small φ₀ (~1e-18) if λ is large, consistent with a
     "Planck-scale" theory. No randomness or fitting is involved — purely
     deterministic.

2) How the Scalar Field Modifies the Tunnel Diode:
   - We assume the tunnel diode has a baseline barrier voltage (V_b ~ 0.05 V).
   - A sinusoidal function in 'voltage' domain simulates the local oscillation
     of the scalar field, multiplied by an exponential decay factor that
     depends on m_val (the "mass" of the field).
   - The barrier is effectively changed from V_b to V_b + scale_factor·φ(Voltage),
     resulting in a different expression for the current (compared to
     the standard quantum baseline).
   - scale_factor (e.g., 1e15) is chosen deterministically to bring tiny φ₀
     values up to the order of ~0.05 V.

3) Code Structure:
   - We load "TunneldiodeTotal.csv" which contains columns: V (voltage) and I (measured current).
   - We define:
       * current_quantum() → the standard Shockley-based "Quantum Baseline"
       * current_scalar() → your Scalar Field model
   - We compute four metrics: RMS, MAE, AIC, BIC, then compare results.
   - Output:
       * A summary in the console,
       * A CSV file with columns [Voltage, I_Exp, I_QM, I_SF],
       * A plot of experimental data vs. both models.

4) Why Omitting MAPE and R²:
   - In certain tunnel diode data, there can be extremely small or nearly zero currents,
     leading to massive percentage errors (MAPE) or very negative R².
   - Hence, to keep results clearer, we only show RMS, MAE, AIC, and BIC here.

--------------------------------------------------------------------------------
"""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from sklearn.metrics import mean_squared_error, mean_absolute_error
import math
import os

# -------------------------------------------------------------------------
# 1) Fundamental Constants (Planck-based approach)
# -------------------------------------------------------------------------
# Per your scalar field theory, these constants define φ₀, m_val, etc.
q    = 1.602e-19      # Elementary charge (C)
k_B  = 1.38e-23       # Boltzmann constant (J/K)
T    = 300            # Temperature (K)

hbar = 1.054571817e-34  # Reduced Planck constant (J*s)
c    = 3.0e8            # Speed of light (m/s)
G    = 6.67430e-11      # Gravitational constant

def planck_length(hb, grav, speed_c):
    """
    Planck length ℓₚ = sqrt( (ħ * G) / c^3 ).
    In your theory, ℓₚ anchors the smallest scale
    from fundamental physics.
    """
    return math.sqrt((hb * grav)/(speed_c**3))

ell_p = planck_length(hbar, G, c)

# Dimensionless lambda (deterministically chosen, not fitted)
lambda_factor = 1.0e10

# m_val = 1 / (lambda * ell_p)
m_val = 1.0/(lambda_factor * ell_p)

# phi_0 = ħ / (lambda * ell_p * c)
phi_0 = hbar/(lambda_factor * ell_p * c)

# Print out the deterministically computed parameters
print("=== Deterministic Scalar Field Params (Planck-based) ===")
print(f"Planck length  = {ell_p:.4e} m")
print(f"lambda_factor  = {lambda_factor:.4e}")
print(f"m_val          = {m_val:.4e} m^-1")
print(f"phi_0          = {phi_0:.4e}")

# -------------------------------------------------------------------------
# 2) Model Definitions
# -------------------------------------------------------------------------
def current_quantum(voltage, I0):
    """
    Quantum Baseline (Shockley-like):
    I(V) = I0 * [ exp(q*V/(k_B*T)) - 1 ]

    This is the standard approximate equation for a diode
    without series/parasitic effects.
    """
    return I0 * (np.exp(q * voltage/(k_B * T)) - 1.0)

def scalar_field_function(voltage):
    """
    Deterministic φ(V):
    φ(V) = φ₀ * sin(2π * V / V_b) * exp(-m_val * (V * L0))

    Explanation:
    - sin(...) introduces a simple oscillatory shape,
      referencing the ratio of (voltage / V_b).
    - exp(-m_val * (V * L0)) simulates an exponential decay
      influenced by m_val (from Planck constants).
    - φ₀ is extremely small ~1e-18, so we rely on a large scale_factor
      later to see any effect in the diode barrier.
    """
    V_b = 0.05   # (V) Typical barrier scale
    L0 = 1.0e-9  # (m) A chosen length scale
    # Sinusoidal part
    sin_part   = np.sin((2.0 * math.pi * voltage) / V_b)
    # Decay part
    decay_part = np.exp(-m_val * voltage * L0)
    return phi_0 * sin_part * decay_part

def current_scalar(voltage, I0):
    """
    Scalar Field Model:
    I(V) = I0 * [ exp(q*V/(k_B*T)) * exp(- V_eff / V_b) - 1 ],
    where V_eff = V_b + scale_factor * φ(V).

    In your theory, the scalar field modifies
    the effective barrier from V_b to V_eff.

    scale_factor is a deterministic constant (e.g. 1e15)
    to scale tiny φ(V) up to the ~0.05 V range.
    """
    V_b = 0.05
    scale_factor = 1.0e15  # Deterministic, not fitted
    phi_val = scalar_field_function(voltage)
    V_eff   = V_b + (scale_factor * phi_val)

    exponent_term = np.exp(q * voltage/(k_B*T)) * np.exp(- V_eff / V_b)
    return I0 * (exponent_term - 1.0)

# -------------------------------------------------------------------------
# 3) Metrics: RMS, MAE, AIC, BIC
# -------------------------------------------------------------------------
def compute_rms(y_true, y_pred):
    """
    Root Mean Square Error (RMS).
    Generally smaller => better fit to data.
    """
    return math.sqrt(mean_squared_error(y_true, y_pred))

def compute_mae(y_true, y_pred):
    """
    Mean Absolute Error (MAE).
    Also smaller => better fit to data.
    """
    return mean_absolute_error(y_true, y_pred)

def compute_aic(y_true, y_pred, num_params):
    """
    Akaike Information Criterion:
    AIC = n * ln(RSS / n) + 2 * k
    Lower AIC => better model (penalizes extra parameters).
    """
    n = len(y_true)
    rss = np.sum((y_true - y_pred)**2)
    return n * math.log(rss / n + 1e-30) + 2 * num_params

def compute_bic(y_true, y_pred, num_params):
    """
    Bayesian Information Criterion:
    BIC = n * ln(RSS / n) + k * ln(n)
    Similar to AIC but penalizes # of parameters more strongly
    for large n.
    """
    n = len(y_true)
    rss = np.sum((y_true - y_pred)**2)
    return n * math.log(rss / n + 1e-30) + num_params * math.log(n)

# -------------------------------------------------------------------------
# 4) Main Routine
# -------------------------------------------------------------------------
def main():
    """
    Main steps of the script:
     1) Load the dataset "TunneldiodeTotal.csv"
        (voltage in 'V', current in 'I').
     2) Compute current predictions from:
        - current_quantum (QM baseline),
        - current_scalar (Scalar Field).
     3) Evaluate RMS, MAE, AIC, and BIC for each model.
     4) Output the results, save them in CSV, and show a plot.
    """
    # Load dataset
    file_path = "TunneldiodeTotal.csv"
    df = pd.read_csv(file_path)
    voltage = df['V'].values
    current_exp = df['I'].values

    # Sufficient for RMS, MAE, AIC, BIC:
    # We'll not filter near-zero points
    # because MAPE and R^2 are omitted.

    I0 = 1.0e-3  # A, saturation current (deterministic)

    # Compute both models
    current_qm = current_quantum(voltage, I0)
    current_sf = current_scalar(voltage, I0)

    # We'll treat each model as having 1 parameter (I0),
    # since phi_0, m_val, scale_factor are fixed from your theory.
    num_params_qm = 1
    num_params_sf = 1

    # RMS & MAE
    rms_qm = compute_rms(current_exp, current_qm)
    rms_sf = compute_rms(current_exp, current_sf)

    mae_qm = compute_mae(current_exp, current_qm)
    mae_sf = compute_mae(current_exp, current_sf)

    # AIC & BIC
    aic_qm = compute_aic(current_exp, current_qm, num_params_qm)
    aic_sf = compute_aic(current_exp, current_sf, num_params_sf)

    bic_qm = compute_bic(current_exp, current_qm, num_params_qm)
    bic_sf = compute_bic(current_exp, current_sf, num_params_sf)

    # Percent improvement in RMS for Scalar Field vs Quantum
    improvement_rms = (1.0 - rms_sf / rms_qm) * 100.0 if rms_qm != 0 else 0.0

    # Print results
    print("\n=== Comparison: Quantum vs. Scalar Field (No MAPE, No R^2) ===")
    print(f"Quantum RMS  : {rms_qm:.5e}")
    print(f"Scalar RMS   : {rms_sf:.5e}")
    print(f"RMS Improvement: {improvement_rms:.2f}%")

    print("----- Other metrics -----")
    print(f"MAE (QM) : {mae_qm:.5e},  MAE (SF): {mae_sf:.5e}")
    print(f"AIC(QM)  : {aic_qm:.3f},   AIC(SF) : {aic_sf:.3f}")
    print(f"BIC(QM)  : {bic_qm:.3f},   BIC(SF) : {bic_sf:.3f}")

    # Save predictions to a CSV
    out_df = pd.DataFrame({
        "Voltage": voltage,
        "I_Exp": current_exp,
        "I_QM": current_qm,
        "I_SF": current_sf
    })
    out_dir = "Output_Detailed"
    os.makedirs(out_dir, exist_ok=True)
    out_path = os.path.join(out_dir, "scalar_vs_quantum_results_noMAPE_R2.csv")
    out_df.to_csv(out_path, index=False)
    print(f"\nDetailed results saved to {out_path}")

    # Plot
    plt.figure(figsize=(10,6))
    plt.plot(voltage, current_exp, 'o-', label="Experimental Data")
    plt.plot(voltage, current_qm, '--', label="Quantum Baseline")
    plt.plot(voltage, current_sf, '-.', label="Scalar Field (Planck-based + sinus)")
    plt.title("Tunnel Diode: Scalar Field vs. QM (No MAPE / No R^2)")
    plt.xlabel("Voltage (V)")
    plt.ylabel("Current (A)")
    plt.grid(True)
    plt.legend()
    plt.show()


if __name__ == "__main__":
    main()
